## 设置精度 多卡 选择
import os
import torch
import time
import logging
import numpy as np
from datasets import Dataset
from torch.utils.data import DataLoader
from transformers import BertTokenizer, BertForMaskedLM, Trainer, TrainingArguments
from pynvml import nvmlInit, nvmlDeviceGetHandleByIndex, nvmlDeviceGetMemoryInfo, nvmlDeviceGetPowerUsage, nvmlShutdown
from tqdm import tqdm
from torch.cuda.amp import autocast, GradScaler  # 【修改】引入混合精度模块

# ----------------------- 配置参数【TODO】 -----------------------
TOKENIZER_PATH = "./BERT_wikitext-2-raw-v1/BERT_wikitext-2-raw-v1/bert_tokenizer"
BERT_MODEL_PATH = "./BERT_wikitext-2-raw-v1/BERT_wikitext-2-raw-v1/bert_model_base"  # 可选base或large
DATASET_PATH = "./BERT_wikitext-2-raw-v1/BERT_wikitext-2-raw-v1/data"
OUTPUT_DIR = "./results"
LOG_DIR = "./logs"
TRAIN_BATCH_SIZE = 32
EVAL_BATCH_SIZE = 32
MAX_STEPS = 200
NUM_EPOCHS = 2
LOGGING_STEPS = 10
SAVE_TOTAL_LIMIT = 1
GPU_COST_PER_HOUR = 1.98  # GPU价格（￥/hr）

# ----------------------- GPU选择与设置【TODO】 -----------------------
PREFERRED_GPU = 0  # 首选GPU编号
USE_MULTIPLE_GPUS = False  # 是否使用多卡训练

if USE_MULTIPLE_GPUS:
    # 若使用多卡训练，则选择所有GPU
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, [PREFERRED_GPU, 1, 2, 3]))
else:
    # 若仅使用单卡训练，则仅选择首选GPU
    os.environ["CUDA_VISIBLE_DEVICES"] = str(PREFERRED_GPU)

# ----------------------- 训练精度选择【TODO】 -----------------------
# 默认精度是单精度FP32。可以选择开启混合精度AMP如下
# 混合精度设置
USE_FP16 = True  # 是否开启混合精度

# 也可以选择强制执行FP16精度。即：USE_FP16设为False
# 然后在调用模型后，model = BertForMaskedLM.from_pretrained(BERT_MODEL_PATH)
# 增加代码：model.half()
# 注意可能导致数值不稳定或溢出。


# ----------------------- 日志设置【TODO】 -----------------------
# 清除所有的日志处理器，避免冲突
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)
logging.basicConfig(filename="training_log_base_mixfp.log", level=logging.INFO, format="%(asctime)s - %(message)s")

# 初始化 GPU 监控
nvmlInit()
gpu_handle = nvmlDeviceGetHandleByIndex(0)

# ----------------------- 数据加载与预处理 -----------------------
def load_local_dataset(file_path):
    with open(file_path, 'r', encoding='utf-8') as f:
        return [{"text": line.strip()} for line in f]

train_data = load_local_dataset(os.path.join(DATASET_PATH, "train.txt"))
eval_data = load_local_dataset(os.path.join(DATASET_PATH, "validation.txt"))

train_dataset = Dataset.from_list(train_data)
eval_dataset = Dataset.from_list(eval_data)

# 分词器与数据处理
tokenizer = BertTokenizer.from_pretrained(TOKENIZER_PATH)

def tokenize_function(examples):
    tokenized = tokenizer(examples["text"], padding="max_length", truncation=True, max_length=128)
    tokenized["labels"] = tokenized["input_ids"].copy()
    return tokenized

train_dataset = train_dataset.map(tokenize_function, batched=True, remove_columns=["text"])
eval_dataset = eval_dataset.map(tokenize_function, batched=True, remove_columns=["text"])

# 自定义 collate_fn
def collate_fn(examples):
    input_ids = torch.stack([torch.tensor(f['input_ids']) for f in examples])
    attention_mask = torch.stack([torch.tensor(f['attention_mask']) for f in examples])
    labels = torch.stack([torch.tensor(f['labels']) for f in examples])
    return {"input_ids": input_ids, "attention_mask": attention_mask, "labels": labels}

train_dataloader = DataLoader(train_dataset, batch_size=TRAIN_BATCH_SIZE, collate_fn=collate_fn)
eval_dataloader = DataLoader(eval_dataset, batch_size=EVAL_BATCH_SIZE, collate_fn=collate_fn)

# ----------------------- 模型加载与训练 -----------------------
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model = BertForMaskedLM.from_pretrained(BERT_MODEL_PATH)

# 判断是否支持多卡
if USE_MULTIPLE_GPUS and torch.cuda.device_count() > 1:
    model = torch.nn.DataParallel(model)

model = model.to(device)

training_args = TrainingArguments(
    output_dir=OUTPUT_DIR,
    eval_strategy="epoch",
    save_strategy="epoch",
    per_device_train_batch_size=TRAIN_BATCH_SIZE,
    logging_dir=LOG_DIR,
    logging_steps=LOGGING_STEPS,
    max_steps=MAX_STEPS,
    num_train_epochs=NUM_EPOCHS,
    save_total_limit=SAVE_TOTAL_LIMIT,
    load_best_model_at_end=True
)

trainer = Trainer(
    model=model,
    args=training_args,
    train_dataset=train_dataset,
    eval_dataset=eval_dataset
)

# 初始化优化器和学习率调度器
trainer.create_optimizer_and_scheduler(num_training_steps=MAX_STEPS)

# 使用GradScaler【修改】
scaler = GradScaler()

# 记录数据列表
throughputs, step_times, mem_usages, power_usages, epoch_times = [], [], [], [], []

# 开始训练
for epoch in tqdm(range(NUM_EPOCHS), desc="Epochs"):
    epoch_loss = 0
    start_time = time.time()
    num_samples = 0

    for step, batch in enumerate(tqdm(train_dataloader, desc="Steps")):
        step_start_time = time.time()
        
        batch = {k: v.to(device) for k, v in batch.items()}
        trainer.optimizer.zero_grad()

        with autocast(enabled=USE_FP16):  # 混合精度训练
            loss = trainer.compute_loss(trainer.model, batch)

        scaler.scale(loss).backward()  # 使用scaler反向传播
        scaler.step(trainer.optimizer)
        scaler.update()

        trainer.lr_scheduler.step()
        
        epoch_loss += loss.item()
        step_time = time.time() - step_start_time
        num_samples += batch["input_ids"].size(0)
        
        throughput = num_samples / (time.time() - start_time)
        mem_info = nvmlDeviceGetMemoryInfo(gpu_handle)
        power_usage = nvmlDeviceGetPowerUsage(gpu_handle) / 1000  # W

        throughputs.append(throughput)
        step_times.append(step_time)
        mem_usages.append(mem_info.used / 1024**2)  # MB
        power_usages.append(power_usage)

    epoch_duration = time.time() - start_time
    epoch_times.append(epoch_duration)
    epoch_loss /= step
    logging.info(f'Epoch {epoch}, Loss: {epoch_loss:.4f}, Throughput: {throughput:.2f} samples/s')

# 关闭 GPU 监控
nvmlShutdown()

# 计算并打印平均指标
avg_throughput = np.mean(throughputs)
avg_step_time = np.mean(step_times)
avg_mem_usage = np.mean(mem_usages)
avg_power_usage = np.mean(power_usages)
avg_epoch_time = np.mean(epoch_times)

logging.info(f"Average Throughput: {avg_throughput:.2f} samples/s")
logging.info(f"Average Step Time: {avg_step_time:.4f}s")
logging.info(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
logging.info(f"Average Power Usage: {avg_power_usage:.2f} W")
logging.info(f"Average Epoch Duration: {avg_epoch_time:.2f}s")

print(f"Average Throughput: {avg_throughput:.2f} samples/s")
print(f"Average Step Time: {avg_step_time:.4f}s")
print(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
print(f"Average Power Usage: {avg_power_usage:.2f} W")
print(f"Average Epoch Duration: {avg_epoch_time:.2f}s")

# 计算 10,000 步迭代成本
gpu_cost_per_hour =  GPU_COST_PER_HOUR  # 每小时 GPU 费用
total_steps = 10000
total_time_hours = (total_steps * avg_step_time) / 3600
total_cost = total_time_hours * gpu_cost_per_hour

logging.info(f"Cost for 10k steps: ${total_cost:.2f}, Total Time: {total_time_hours:.2f} hours")
print(f"Cost for 10k steps: ${total_cost:.2f}, Total Time: {total_time_hours:.2f} hours")
